//-------------------------------------------------------------------
//   FSConnect SDK
//   Copyright  2005 Russel Dirks.  All rights reserved.
//   Email : russdirks@yahoo.ca
//-------------------------------------------------------------------
#include <windows.h>
#include <stdio.h>
#include "Tokens.h"
#include "FSConnect.h"

HWND FSCHWnd = 0;		// window handle to FSConnect
int FSCMsg;				// window message sent to FSConnect

//-------------------------------------------------------------------
// Prototypes
//-------------------------------------------------------------------
void HexToBin(char *pHexStr, char *pBin, int nBytes);
UINT HexToUINT(char *p);
void BinToHex(char *pDest, char *pSrc, int nBytes);

//-------------------------------------------------------------------
// InitFSConnect  - initialization routine
//
// Return Value: zero = failure, non-zero = success
//-------------------------------------------------------------------
BOOL InitFSConnect() {

	if ((FSCMsg != 0) && (FSCHWnd != 0))	// FSConnect has already been initialized
		return TRUE;

   FSCMsg = RegisterWindowMessage(WM_FSCONNECT);
   FSCHWnd = FindWindow(FSC_Class, NULL);   // locate FSConnect
   return (FSCMsg != 0) && (FSCHWnd != 0);
}

//-------------------------------------------------------------------
// The next four functions are high level functions to get/set latitude &
// longitude.  They all return 0 for success, non-zero for error.  See
// defined constants in FSConnect.h for meaning of error values
//-------------------------------------------------------------------

int fscGetLatitude(double *pLat) {
	UNIVERSAL_VAR uv;
	VAR_TYPE vt;
	int res;

	res = fscGetVar(PLANE_LATITUDE, &uv, &vt);
	if (res == 0) {
		*pLat = uv.n / 40007000.0 * 360.0;
	} 
	return res;
}

int fscGetLongitude(double *pLon) {
	UNIVERSAL_VAR uv;
	VAR_TYPE vt;
	int res;

   res = fscGetVar(PLANE_LONGITUDE, &uv, &vt);
	if (res == 0) {
		*pLon = uv.n / 281474976710656.0 * 360.0;
	} 
	return res;
}

int fscSetLatitude(double Lat) {
	__int64 iLat;
	char pBuff[MaxAtomStr];
	
	iLat = (__int64)((Lat / 90.0) * (10001750.0 * 65536.0 * 65536.0));
	BinToHex(pBuff, (char *) &iLat, 8);
	return fscSetVar(PLANE_LATITUDE, pBuff);
}

int fscSetLongitude(double Lon) {
	__int64 iLon;
	char pBuff[MaxAtomStr];
	
	iLon = (__int64)((Lon / 360.0) * (65536.0 * 65536.0 * 65536.0 * 65536.0));
	BinToHex(pBuff, (char *) &iLon, 8);
	return fscSetVar(PLANE_LONGITUDE, pBuff);
}


//-------------------------------------------------------------------
// fscGetVar  - request data from FSConnect
//
//     Token [in] - data element we want from Flight Simulator
//     UVar [out] - buffer to hold requested data value
//     VarType [out] - var type as reported by Flight Simulator 
//
// Return Value: 0 if success, non-zero for error (see defined constants)
//
// COMMENTS:
// You can use the value returned in VarType if you are having trouble interpreting the data 
// that comes back.  This is set by FS, and may be more accurate than the SDK.
//-------------------------------------------------------------------
int fscGetVar(GAUGE_TOKEN Token, UNIVERSAL_VAR *UVar, VAR_TYPE *VarType) {

	int iAtom, chars;
	char buff[MaxAtomStr];

	if (!InitFSConnect()) return errNoFSC;
	// send request.  FSConnect sends back a handle to an atom, which
	// contains the data we want
	iAtom = SendMessage(FSCHWnd, FSCMsg, GETVAR, Token);
	if (iAtom > 0)		// string atoms are always negative
		return iAtom;   // iAtom contains error code
	else if (iAtom == 0)
		return errNoFSC;
	chars = GlobalGetAtomName(iAtom, buff, MaxAtomStr);		// retrieve data
	GlobalDeleteAtom(iAtom);
	*VarType = (VAR_TYPE) HexToUINT(buff);	// first 4 bytes is vartype
	HexToBin(buff + 8, (char *)UVar, 8); // next 8 bytes is universal var
	return 0;
}


//-------------------------------------------------------------------
// fscGetString  - request data from FSConnect
//
//     Token [in] - data element we want from Flight Simulator
//     pStr [out] - pointer to a buffer to hold string data returned
//
// Return Value: 0 if success, non-zero for error (see defined constants above)
//
// COMMENTS:
// You will notice that certain data elements listed in TokenVar.doc show a "P"
// in the Universal_Var column. The "P" stands for pointer. Some examples are
// VOR1_IDENTITY and GPS_WP_STRING. What that means is the data stored in the
// Universal_Var buffer is a pointer to the data, not the data itself.
// If you were to use the fscGetVar function above, you would only be getting the
// pointer, which doesn't do you a whole lot of good, since you cannot access the
// memory space of another process.  This function will give you the actual string.
// Only call this function with "P" elements.
//-------------------------------------------------------------------
int fscGetString(GAUGE_TOKEN Token, char *pStr) {
   int iAtom;
   
	if (!InitFSConnect()) return errNoFSC;
	pStr[0] = 0;	// initialize to empty string
   // send request
   iAtom = SendMessage(FSCHWnd, FSCMsg, GETSTRING, Token);
	if (iAtom > 0)		// string atoms are always negative
		return iAtom;   // iAtom contains error code
	else if (iAtom == 0)
		return errNoFSC;
   GlobalGetAtomName(iAtom, pStr, MaxAtomStr); // retrieve data
   GlobalDeleteAtom(iAtom);
   return 0;
}


//-------------------------------------------------------------------
// fscSetVar  - send data to be written into Flight Simulator.  Currently
//              only latitude and longitude are supported. (FS must be in
//				  pause or slew mode)
//
//     Token [in]    - data element ID
//     pVarData [in] - data to write, in hexadecimal string format,
//                     "memory" byte order
//
// Return Value: 0 if success, non-zero for error (see defined constants above)
//
//-------------------------------------------------------------------
int fscSetVar(GAUGE_TOKEN Token, char *pVarData) {
   int iAtom, res;
	char sAtom[MaxAtomStr];
	   
	if (!InitFSConnect()) return errNoFSC;
	sprintf(sAtom,"%.8X",Token);
	strcat(sAtom, pVarData);
   iAtom = GlobalAddAtom(sAtom);
   res = SendMessage(FSCHWnd, FSCMsg, SETVAR, iAtom);
   GlobalDeleteAtom(iAtom);
	return res;
}

//-------------------------------------------------------------------
// fscGetVersions  - returns FSConnect version, FS Exe name, and FS Version
//
//     FSCVersion [out] - version of FSConnect.dll (5 character string)
//     FSExeName [out] - exe name of FS (eg, "fs2002.exe")
//     FSVerHi [out] - FS Version number. 
//     FSVerLo [out] - FS Version number.
//
// Return Value: 0 = success, non-zero = error
//-------------------------------------------------------------------
int fscGetVersions(char *FSCVersion, char *FSExeName, UINT *FSVerHi, UINT *FSVerLo) {
   int iAtom;
	char buff[MaxAtomStr];

	if (!InitFSConnect()) return errNoFSC;
   iAtom = SendMessage(FSCHWnd, FSCMsg, GETVERSIONS, 0);
	if (iAtom > 0)		// string atoms are always negative
		return iAtom;   // iAtom contains error code
	else if (iAtom == 0)
		return errNoFSC;
   GlobalGetAtomName(iAtom, buff, MaxAtomStr); // retrieve data
   GlobalDeleteAtom(iAtom);
	strncpy(FSCVersion, buff, 5);
	FSCVersion[5] = 0;	// terminate the string
	*FSVerHi = HexToUINT(buff + 5);
	*FSVerLo = HexToUINT(buff + 13);
	strcpy(FSExeName, buff + 21);
	return 0;
}

//-------------------------------------------------------------------
// fscMemRead  - instruct FSConnect to read and return data from a
//               specified address from with Flight Simulator's memory
//               space.
//
//     Addr [in]    - address to read from
//     nBytes [in]  - number of bytes to read (max 127)
//     pszMem [out] - pointer to buffer to hold data returned.  Data comes
//	                 back in hexadecimal string format, in "memory" byte order
//
// Return Value: zero if success, non-zero if error
//
//-------------------------------------------------------------------
int fscMemRead(PVOID addr, int nBytes, LPSTR pszMem) {
	char sAtom[MaxAtomStr];
	int iAtomIn, iAtomOut;
   
	if (!InitFSConnect()) return errNoFSC;
	if (nBytes > (MaxAtomStr / 2)) return errInvalidData;
	sprintf(sAtom,"%.8X%.8X",addr,nBytes);
	iAtomIn = GlobalAddAtom(sAtom);
   iAtomOut = SendMessage(FSCHWnd, FSCMsg, MEMREAD, iAtomIn);
	if (iAtomOut > 0)		// string atoms are always negative
		return iAtomOut;   // iAtom contains error code
	else if (iAtomOut == 0)
		return errNoFSC;
	GlobalGetAtomName(iAtomOut, pszMem, nBytes * 2 + 1);
   GlobalDeleteAtom(iAtomIn);
   GlobalDeleteAtom(iAtomOut);
   return 0;
}

//-------------------------------------------------------------------
// fscMemWrite  - instruct FSConnect to write data to a specific
//                address in Flight Simulator's memory space.
//
//     Addr [in]    - address to write to 
//     pszData [in] - data to write, in hexadecimal string format 
//	                 (max 247 chars), in "memory" byte order
//
// Return Value: zero if success, non-zero if error
//
//-------------------------------------------------------------------
int fscMemWrite(PVOID Addr, LPSTR pszData) {
	char sAtom[MaxAtomStr];
	int iAtom, res;

	if (!InitFSConnect()) return errNoFSC;
   if (strlen(pszData) > (MaxAtomStr - 8)) return errInvalidData;
	sprintf(sAtom, "%.8X%s", Addr, pszData);
   iAtom = GlobalAddAtom(sAtom);
   res = SendMessage(FSCHWnd, FSCMsg, MEMWRITE, iAtom);
   GlobalDeleteAtom(iAtom);
	return res;
}

//-------------------------------------------------------------------
// fscSearchMemRange  - instruct FSConnect to search a range of addresses
//                      in Flight Simulator's memory space for specified data
//
//     StartAddr [in] - Start address of search range
//     EndAddr [in]   - End address of search range
//     SrchData [in]  - Data to search for, in hexadecimal string format,
//                      in "memory" byte order. Should be 2, 4, 8, or 16
//                      bytes long, corresponding to 1, 2, 4, or 8 bytes
//                      of data.
//     FoundAt [out]  - address of data, if found
//
// Return Value: 0 if data found, otherwise error value
//
// Remarks: searching is very quick (1MB takes about a second) provided you limit
// your searching to allocated memory.  If, however, you specify memory
// that has not been allocated, each read will generate an access violation,
// which will slow the search significantly!!
//-------------------------------------------------------------------
int fscSearchMemRange(PVOID StartAddr, PVOID EndAddr, LPCSTR SrchData, PVOID *FoundAt) {
   int iAtomIn, iAtomOut;
	char sAtom[MaxAtomStr], szFoundAt[10];
   
	if (!InitFSConnect()) return errNoFSC;
	sprintf(sAtom, "%.8X%.8X%s", StartAddr, EndAddr, SrchData);
   iAtomIn = GlobalAddAtom(sAtom);
   iAtomOut = SendMessage(FSCHWnd, FSCMsg, SEARCHMEMRNG, iAtomIn);
	if (iAtomOut > 0)		// string atoms are always negative
		return iAtomOut;   // iAtom contains error code
	else if (iAtomOut == 0)
		return errNoFSC;
   GlobalGetAtomName(iAtomOut, szFoundAt, 10);

   *FoundAt = (PVOID)HexToUINT(szFoundAt);
   GlobalDeleteAtom(iAtomIn);
   GlobalDeleteAtom(iAtomOut);
   return 0;
}

//-------------------------------------------------------------------
// fscSearchMemAll  - instruct FSConnect to search all allocated memory
//                    in Flight Simulator's memory space for specified data
//
//     SrchData [in]  - Data to search for, in hexadecimal string format,
//                      in "memory" byte order. Should be 2, 4, 8, or 16
//                      bytes long, corresponding to 1, 2, 4, or 8 bytes
//                      of data.
//     ResultsFile [in] - name of file to receive search results
//
// Return Value: 0 if search completed, otherwise error value
//
// Remarks: this search takes about 8 seconds on my PIII-600
//-------------------------------------------------------------------
int fscSearchMemAll(LPCSTR SrchData, LPCSTR ResultsFile) {
   int iAtom, res;
	char sAtom[MaxAtomStr];
   
	if (!InitFSConnect()) return errNoFSC;
	sprintf(sAtom, "%.2X%s%s", strlen(SrchData) / 2, SrchData, ResultsFile);
   iAtom = GlobalAddAtom(sAtom);
   res = SendMessage(FSCHWnd, FSCMsg, SEARCHMEMALL, iAtom);
   GlobalDeleteAtom(iAtom);
	return res;
}

//-------------------------------------------------------------------
// fscTriggerKeyEvent  - This function exposes the function 
//     trigger_key_event. See the Panels & Guages SDK for complete 
//     information.  In brief, this function allows you to control
//     the user aircraft, and certain aspects of the sim.  For instance,
//     you can manipulate the flight controls, or set switches on 
//     the panel.
//
//     event_id [in]  - Event identifier.  See EventIDs.doc in the Panels
//                      & Guages SDK for complete list.
//     value [in]     - value to set
//
// Return Value: this function returns whatever trigger_key_event 
//     returns, which seems to be zero for success.
//-------------------------------------------------------------------
int fscTriggerKeyEvent(UINT event_id, UINT value) {
   int iAtom, res;
	char sAtom[MaxAtomStr];
   
	if (!InitFSConnect()) return errNoFSC;
	sprintf(sAtom, "%.8X%.8X", event_id, value);
   iAtom = GlobalAddAtom(sAtom);
   res = SendMessage(FSCHWnd, FSCMsg, TRIGGERKEY, iAtom);
   GlobalDeleteAtom(iAtom);
	return res;
}

//--------------------------------------------------------------------------------
// fscGetXMLVar  - This function exposes the execute_calculator_code function.  
//     There is no mention of this function in the SDK, however various people on 
//     Avsim and elsewhere have figured out you can pass it an XML expression and it 
//     will return the result.
//
//     XML_Expr [in]   - XML expression
//     fvalue [in/out] - pointer to double to receive return value. If you're not
//                       interested in this value, pass in 0
//     ivalue [in/out] - pointer to integer to receive return value.  If you're not 
//                       interested in this value, pass in 0
//     svalue [in/out] - pointer to string buffer, at least 255 bytes in size, to 
//                       receive return value.  If you're not interested in this value, 
//                       pass in 0
//
// Return Value: this function returns whatever execute_calculator_code 
//     returns, which seems to be 1 for success.
//--------------------------------------------------------------------------------
int fscGetXMLVar(PCHAR XML_Expr, double* fvalue, int* ivalue, char* svalue) {
	char sAtom[MaxAtomStr];
	int iAtomIn, iAtomOut, res;
   
	if (!InitFSConnect()) return errNoFSC;
	iAtomIn = GlobalAddAtom(XML_Expr);
   iAtomOut = SendMessage(FSCHWnd, FSCMsg, GETXMLVAR, iAtomIn);
	if (iAtomOut > 0)		
		return iAtomOut;  
	else if (iAtomOut == 0)
		return errNoFSC;
	GlobalGetAtomName(iAtomOut, sAtom, MaxAtomStr);
   GlobalDeleteAtom(iAtomIn);
   GlobalDeleteAtom(iAtomOut);
	// unpack the data
	res = HexToUINT(sAtom);
	if (fvalue) 
		HexToBin(sAtom + 8, (char *)fvalue, 8);
	if (ivalue)
		*ivalue = HexToUINT(sAtom + 24);
	if (svalue)
		strcpy(svalue, sAtom + 32);
   return res;
}

//-------------------------------------------------------------------
// Miscellaneous functions
//-------------------------------------------------------------------

// take hex characters from pHexStr, convert them to binary, and store in pBin
void HexToBin(char *pHexStr, char *pBin, int nBytes) {
	int i, sb;
	char *temp;

	for (i=0; i<nBytes;i++, pHexStr += 2) {
		sb = pHexStr[2];
		pHexStr[2] = 0;
		pBin[i] = (BYTE)strtoul(pHexStr,&temp,16);
		pHexStr[2] = sb;
	}
}

UINT HexToUINT(char *p) {
	int savebyte;
	long res;
	char *temp;

	savebyte = p[8];
	p[8] = 0;
	res = strtoul(p,&temp,16);
	p[8] = savebyte;
	return res;
}

void BinToHex(char *pDest, char *pSrc, int nBytes) {
	int i;

	for (i=0; i < nBytes; i++) {
		sprintf(pDest + (i * 2),"%.2X",(unsigned char)pSrc[i]);
	}
	pDest[i*2] = 0;
}

