//----------------------------------------------------------------------------
//		- Author : Rudolf Schider  (www.schider.com)
//
//    - Modifications and additions : Russel Dirks
//----------------------------------------------------------------------------

unit FSConnect;

  {$VARSTRINGCHECKS OFF}

interface

uses
  Tokens;

const
  errOk              = 0;
  //error messages from this unit
  errNoFSC           = 1;  //unable to communicate with FSConnect
  errBadVariant      = 2;  //variant type not implemented
  errBadAtom				 = 3;
  //error messages returned by FSConnect.dll
  errNoFSWnd         = 4;  // couldn't find FS (not used)
  errRegWndClass     = 5;  // an error occurred registering the FSConnect window class
  errCreateWnd       = 6;  // an error occurred creating the FSConnect window
  errNotImplemented  = 7;  // the command you requested has not been implemented
  errInvalidData     = 8;  // the command you sent contained invalid data
  errAccessViolation = 9;  // a read/write you requested generated an access violation
  errNotFound        = 11; // SearchMem function did not find specified data
	errFSClosing			 = 16; // FS is closing

type
	//  The following type definition is used when you call fscGetVar to specify what
  // data type you are expecting back.  You determine this from TokenVar.doc
  // Use the following mapping:
  TVarType = (vtDouble,     // "N"
  						vtBCD,				// "D"
              vtOctal,			// "O"
              vtBoolean,		// "B"
              vtFlags,			// "F"
              vtEnum);			// "E"

  // Note that for "P" elements, you call fscGetString, not fscGetVar

//-----------------------------------------------------------------------------

function fscGetVersions (var FSCVersion, FSExeName : string;
                         var FSVerHi, FSVerLo : cardinal) : integer;
function fscGetString (Token : GAUGE_TOKEN; var S : string) : integer;
function fscGetVar (Token : GAUGE_TOKEN; VarType : TVarType; var VarValue : Variant) : integer;
function fscSetVar (Token : integer; S : string) : integer;

function  fscGetLatitude(Var Lat : double) : integer;
function  fscGetLongitude(Var Lon : double) : integer;

(*
procedure fscSetLatitude (lat : double);
procedure fscSetLongitude (lon : double);
*)

function fscSearchMemRange (StartAddr, EndAddr : Longint;
                            SearchStr : string; var FoundAt : Longint) : integer;
function fscSearchMemAll (SrchData, ResultsFile : string) : integer;
function fscTriggerKeyEvent(event_id, value : cardinal) : integer;
function fscGetXMLVar(XML_Expr : string; var fvalue : double; var ivalue : integer;
											var svalue : string) : integer;

//-----------------------------------------------------------------------------

implementation

uses
  Classes,
  SysUtils,
  {$IFDEF VER140} Variants, {$ENDIF}
  Windows;

//-----------------------------------------------------------------------------

type
  TAtom = smallint;   //16 bit signed integer

const
  //these are commands sent to FSConnect
  GETVAR       = 1;
  GETSTRING    = 2;
  GETPBIN      = 3;        //  not implemented
  MEMREAD      = 4;
  MEMWRITE     = 5;
  SETVAR       = 6;        //  only Lat/Lon supported at this time
  SEARCHMEMRNG = 8;
  SEARCHMEMALL = 9;
  GETVERSIONS  = 10;
	GETXMLVAR		 = 14;
	TRIGGERKEY	 = 15;

  WM_FSCONNECT = 'WM_FSCONNECT';
  FSC_Class    = 'FSCONNECT8';
  MaxAtomStr    = 255;

var
  FSCHWnd : integer = 0;   //window handle to FSConnect
  FSCMsg  : integer = 0;   //window message sent to FSConnect

//-------------------------------------------------------------------
// InitFSConnect  - initialization routine
//
// Return Value: zero = failure, non-zero = success
//-------------------------------------------------------------------
function InitFSConnect : Boolean;
begin
	if (FSCMsg <> 0) and (FSCHWnd <> 0) then
  	result := True	// initialization has already been performed
  else begin
		FSCMsg := RegisterWindowMessage(WM_FSCONNECT);
		FSCHWnd := FindWindowA(FSC_Class,Nil);
	  result := (FSCMsg <> 0) and (FSCHWnd <> 0);
  end;
end;

//-------------------------------------------------------------------
// The next two functions are high level functions to get latitude &
// longitude.  They return 0 for success, non-zero for error.  See
// defined constants above for meaning of error values
//-------------------------------------------------------------------
function fscGetLatitude(Var Lat : double) : integer;
var v : variant;
begin
  result := fscGetVar (PLANE_LATITUDE, vtDouble, v);
  if result = 0 then
    Lat := v / 40007000 * 360;
end;

//-----------------------------------------------------------------------------

function fscGetLongitude(Var Lon : double) : integer;
var v : variant;
begin
  result := fscGetVar (PLANE_LONGITUDE, vtDouble, v);
  if result = 0 then
    Lon := v / 281474976710656.0 * 360;
end;

//-------------------------------------------------------------------
// fscGetVar  - request data from FSConnect
//
//     Token [in] - data element we want from Flight Simulator
//     VarType [in] - data type to return
//     VarValue [out] - data requested
//
// Return Value: 0 if success, non-zero for error (see defined constants)
//
function fscGetVar(Token : GAUGE_TOKEN; VarType : TVarType; var VarValue : Variant) : integer;
var
  iAtom  : TAtom;
  res, i : integer;
  L : Longint;
  S : string;
  buffer : array [0..MaxAtomStr-1] of char;
  StrHex : Array [0..MaxAtomStr] of char;
  D : double;
begin
  if not InitFSConnect then
  	result := errNoFSC
  else begin
    iAtom := SendMessage (FSCHWnd, FSCMsg, GETVAR, Longint (Token));
    if iAtom > 0 then		// atoms are always negative, so +ve number is an error
    	result := iAtom
    else if iAtom = 0 then
    	result := errNoFSC
    else begin
      res := GlobalGetAtomNameA (iAtom, StrHex, MaxAtomStr);
      if res = 0 then
      	result := errBadAtom
      else begin
	      S := string (StrHex);
  	    GlobalDeleteAtom (iAtom);
    	  result := 0;
      end;
    end;
  end;

  if result <> 0 then exit;

  // now unpack the string

  {if you have a later version of Delphi, you can use the builtin
   function : "HexToBin (PChar (S), Buffer, MaxAtomStr);" and delete the
   following for loop }
  for i := 1 to Length (S) div 2 do
  	buffer [i-1] := char (StrToInt ('$' + copy (S, i*2-1, 2)));

  case VarType of
    vtDouble: begin
    		Move (buffer [4], D, 8);
      	VarValue := D;
	      end;
    vtBoolean: begin
        Move (buffer [4], D, 8);
        VarValue := D <> 0;
        end;
    vtBCD, vtOctal:	begin		// Binary Coded Decimal, Octal
       	Move (buffer [4], L, 4);
        VarValue := IntToHex (L, 4);
        end;
    vtFlags, vtEnum: begin
       	Move (buffer [4], L, 4);
        VarValue := L;
        end;
  end;
  result := 0;
end;

//-------------------------------------------------------------------
// fscGetString  - request data from FSConnect
//
//     Token [in] - data element we want from Flight Simulator
//     S [out] - returned data
//
// Return Value: 0 if success, non-zero for error (see defined constants above)
//
// COMMENTS:
// Only use this function for "P" items, as per TokenVar.doc
//-------------------------------------------------------------------
function fscGetString (Token : GAUGE_TOKEN; var S : string) : integer;
var
  iAtom  : TAtom;
  StrHex : Array [0..MaxAtomStr] of char;
begin
  if InitFSConnect then begin
    iAtom := SendMessage (FSCHWnd, FSCMsg, GETSTRING, Integer(Token));
    if iAtom < 0 then begin
      asm nop end;
      GlobalGetAtomNameA (iAtom, StrHex, MaxAtomStr);
      s := string (StrHex);
      GlobalDeleteAtom (iAtom);
      result := errOk;
    end else
      result := iAtom;
  end else
    result := errNoFSC;
end;

//-----------------------------------------------------------------------------

function fscSetVar (Token : integer; S : string) : integer;
var
  iAtom : TAtom;
  StrAtom : string;
begin
  if InitFSConnect then begin
    StrAtom := IntToHex (Token, 8) + S;
    iAtom := GlobalAddAtomA (PChar (StrAtom));
    fscSetVar := SendMessage (FSCHWnd, FSCMsg, SETVAR, iAtom);
    GlobalDeleteAtom (iAtom);
  end else
    result := errNoFSC;
end;

//-------------------------------------------------------------------
// fscMemRead  - instruct FSConnect to read and return data from a
//               specified address from with Flight Simulator's memory
//               space.
//
//     Addr [in]    - address to read from
//     nBytes [in]  - number of bytes to read (max 127)
//     S [out] 			- string to hold data returned.  Data comes
//	                 back in hexadecimal string format
//
// Return Value: zero if success, non-zero if error
//
//-------------------------------------------------------------------
function fscMemRead (Addr : Longint; nBytes : byte; var S : string) : integer;
var
  iAtom : TAtom;
  StrAtom : string;
  iAtomOut : Longint;
  iAtomIn : longint;
  Buffer : array [0..MaxAtomStr] of char;
  chars : Longint;
begin
  S := '';
  if InitFSConnect then begin
    if nBytes <= (MaxAtomStr div 2) then begin
      StrAtom := IntToHex (Addr, 8) + IntToHex (nBytes, 8);
      iAtom := GlobalAddAtomA (PChar (StrAtom));
      iAtomOut := SendMessage (FSCHWnd, FSCMsg, MEMREAD, iAtomIn);
      //string atoms are always negative
      if iAtomOut < 0 then begin
        //retrieve data
        chars := GlobalGetAtomNameA (iAtomOut, Buffer, MaxAtomStr);
        //copy from buffer to result string
        Move (Buffer, S[1], chars);
        result := errOk;
      end else
        result := iAtomOut;
      GlobalDeleteAtom (iAtomIn);
      GlobalDeleteAtom (iAtomOut);
    end else
      result := errInvalidData;
  end else
    result := errNoFSC;
end;

//-------------------------------------------------------------------
// fscMemWrite  - instruct FSConnect to write data to a specific
//                address in Flight Simulator's memory space.
//
//     Addr [in]    - address to write to 
//     HexStr [in] 	- data to write, in hexadecimal string format
//	                 (max 247 chars)
//
// Return Value: zero if success, non-zero if error
//
//-------------------------------------------------------------------
function fscMemWrite (Addr : Longint; HexStr : string) : integer;
var
  iAtom : TAtom;
  StrAtom : string;
begin
  if InitFSConnect then begin
    if length (HexStr) < MaxAtomStr-8 then begin
      StrAtom := IntToHex (Addr, 8) + HexStr;
      iAtom := GlobalAddAtomA (PChar (StrAtom));
      result := SendMessage (FSCHWnd, FSCMsg, MEMWRITE, iAtom);
      GlobalDeleteAtom (iAtom);
    end else
      result := errInvalidData
  end else
    result := errNoFSC;
end;

//-------------------------------------------------------------------
// fscSearchMemRange  - instruct FSConnect to search a range of addresses
//                      in Flight Simulator's memory space for specified data
//
//     StartAddr [in] - Start address of search range
//     EndAddr [in]   - End address of search range
//     SearchStr [in] - Data to search for, in hexadecimal string format,
//                      in "memory" byte order. Should be 2, 4, 8, or 16
//                      bytes long, corresponding to 1, 2, 4, or 8 bytes
//                      of data.
//     FoundAt [out]  - address of data, if found
//
// Return Value: 0 if data found, otherwise error value
//
// Remarks: searching is very quick (1MB takes about a second) provided you limit
// your searching to allocated memory.  If, however, you specify memory
// that has not been allocated, each read will generate an access violation,
// which will slow the search significantly!!
//-------------------------------------------------------------------
function fscSearchMemRange (StartAddr, EndAddr : Longint;
                            SearchStr : string; var FoundAt : Longint) : integer;
var
  StrAtom : string;
  l : integer;
  iAtomIn : TAtom;
  iAtomOut : TAtom;
  Buffer : array [0..10] of char;
  chars : Longint;
begin
  FoundAt := 0;
  if InitFSConnect then begin
    l := length (SearchStr);
    StrAtom := IntToHex (StartAddr, 8) + IntToHex (EndAddr, 8) +
               IntToHex (l div 2, 2) + SearchStr;
    iAtomIn := GlobalAddAtomA (PChar (StrAtom));
    iAtomOut := SendMessage (FSCHWnd, FSCMsg, SEARCHMEMRNG, iAtomIn);
    if iAtomOut > 0 then
      result := iAtomOut
    else if iAtomOut = 0 then
      result := errNoFSC
    else begin
      chars := GlobalGetAtomName (iAtomOut, Buffer, 10);
      if chars >= 8 then
        FoundAt := StrToInt ('$' + Buffer);
    end;
    GlobalDeleteAtom (iAtomIn);
    GlobalDeleteAtom (iAtomOut);
  end else
    result := errNoFSC;
end;

//-------------------------------------------------------------------
// fscSearchMemAll  - instruct FSConnect to search all allocated memory
//                    in Flight Simulator's memory space for specified data
//
//     SrchData [in]  - Data to search for, in hexadecimal string format,
//                      in "memory" byte order. Should be 2, 4, 8, or 16
//                      bytes long, corresponding to 1, 2, 4, or 8 bytes
//                      of data.
//     ResultsFile [in] - name of file to receive search results
//
// Return Value: 0 if search completed, otherwise error value
//-------------------------------------------------------------------
function fscSearchMemAll (SrchData, ResultsFile : string) : integer;
var
  sAtom : string;
  iAtom : TAtom;
begin
  if InitFSConnect then begin
    sAtom := IntToHex (length (SrchData) div 2, 2) + SrchData + ResultsFile;
    iAtom := GlobalAddAtomA (PChar (sAtom));
    result := SendMessage (FSCHWnd, FSCMsg, SEARCHMEMALL, iAtom);
    GlobalDeleteAtom (iAtom);
  end else
    result := errNoFSC;
end;

//-------------------------------------------------------------------
// fscGetVersions  - returns FSConnect version, FS Exe name, and FS Version
//
//     FSCVersion [out] - version of FSConnect.dll (5 character string)
//     FSExeName [out] - exe name of FS (eg, "fs2002.exe")
//     FSVerHi [out] - FS Version number. 
//     FSVerLo [out] - FS Version number.
//
// Return Value: 0 = success, non-zero = error
//-------------------------------------------------------------------
function fscGetVersions (var FSCVersion, FSExeName : string;
                         var FSVerHi, FSVerLo : cardinal) : integer;
var
  iAtom  : LRESULT;
  StrHex : Array [0..MaxAtomStr] of char;
  s : string;
  chars  : UINT;
begin
  if InitFSConnect then begin
    result := errOk;
    iAtom := SendMessage (FSCHWnd, FSCMsg, GETVERSIONS, 0);
    if iAtom < 0 then begin
      chars := GlobalGetAtomNameA (iAtom, StrHex, MaxAtomStr);
      s := string (StrHex);
      GlobalDeleteAtom (iAtom);
      FSCVersion := copy (s, 1, 5);
      FSVerHi    := StrToInt ('$' + copy (s, 6, 8));
      FSVerLo    := StrToInt ('$' + copy (s, 14, 8));
      FSExeName  := copy (s, 22, chars-21);
    end else
      result := iAtom;
  end else
    result := errNoFSC;
end;

//-------------------------------------------------------------------
// fscTriggerKeyEvent  - This function exposes the function 
//     trigger_key_event. See the Panels & Guages SDK for complete 
//     information.  In brief, this function allows you to control
//     the user aircraft, and certain aspects of the sim.  For instance,
//     you can manipulate the flight controls, or set switches on 
//     the panel.
//
//     event_id [in]  - Event identifier.  See EventIDs.doc in the Panels
//                      & Guages SDK for complete list.
//     value [in]     - value to set
//
// Return Value: this function returns whatever trigger_key_event 
//     returns, which seems to be zero for success.
//-------------------------------------------------------------------
function fscTriggerKeyEvent(event_id, value : cardinal) : integer;
var
	iAtom : TAtom;
  res : integer;
	sAtom : string;
begin
	if InitFSConnect then begin
  	sAtom := IntToHex(event_id, 8) + IntToHex(value, 8);
    iAtom := GlobalAddAtomA(PChar(sAtom));
    result := SendMessage(FSCHWnd, FSCMsg, TRIGGERKEY, iAtom);
		GlobalDeleteAtom(iAtom);
	end else
  	result := errNoFSC;
end;

//--------------------------------------------------------------------------------
// fscGetXMLVar  - This function exposes the execute_calculator_code function.  
//     There is no documentation for this function in the P&G SDK, however various people on 
//     Avsim and elsewhere have figured out you can pass it an XML expression and it 
//     will return the result.
//
//     XML_Expr [in]   - XML expression
//     fvalue [in/out] - return value in double format, if applicable.
//     ivalue [in/out] - return value in integer format, if applicable.
//     svalue [in/out] - return value in string format, if applicable. 
//
// Return Value: this function returns whatever execute_calculator_code 
//     returns, which seems to be 1 for success.
//--------------------------------------------------------------------------------
function fscGetXMLVar(XML_Expr : string; var fvalue : double; var ivalue : integer;
											var svalue : string) : integer;
var
	sAtom, s : string;
	iAtomIn, iAtomOut : TAtom;
  res, i : integer;
  buffer : Array [0..MaxAtomStr-1] of char;
begin
	if not InitFSConnect then
  	result := errNoFSC
  else begin
		iAtomIn := GlobalAddAtomA(PChar(XML_Expr));
   	iAtomOut := SendMessage(FSCHWnd, FSCMsg, GETXMLVAR, iAtomIn);
		if (iAtomOut > 0) then
			result := iAtomOut
		else if (iAtomOut = 0) then
			result := errNoFSC
    else begin
			GlobalGetAtomName(iAtomOut, buffer, MaxAtomStr);
      s := string (buffer);
   		GlobalDeleteAtom(iAtomIn);
   		GlobalDeleteAtom(iAtomOut);
			// unpack the data
			result := StrToInt ('$' + copy (s, 1, 8));
      // get the double value
	    for i := 1 to 8 do
  	  	buffer [i-1] := char (StrToInt ('$' + copy (s, i*2-1+8, 2)));
   		Move (buffer [0], fvalue, 8);
			ivalue := StrToInt ('$' + copy (s, 25, 8));
      svalue := copy(s, 33, 999);
		end;
  end;
end;


end.


